@@ -6,13 +6,81 @@
 #include <stdlib.h>
 #include <stdarg.h>
 
+#ifdef __APPLE__
+#include <pthread.h>
+
+#include <AvailabilityMacros.h>
+#include <TargetConditionals.h>
+
+/* TARGET_OS_OSX is not defined on older macOS, but is always defined
+   to 0 for Apple mobile targets. */
+#if !defined(TARGET_OS_OSX) || TARGET_OS_OSX
+#include <libkern/OSCacheControl.h>
+#endif
+#endif
+
+#if defined(HAVE_CODEMEM_VIRTUALALLOC)
+#include <windows.h>
+  #ifdef ORC_WINAPI_ONLY_APP
+    #define _virtualprotect VirtualProtectFromApp
+  #else
+    #define _virtualprotect VirtualProtect
+  #endif
+#endif
+
 #include <orc/orcprogram.h>
 #include <orc/orcdebug.h>
+#include <orc/orcinternal.h>
+#include <orc/orcutils-private.h>
 
 #ifdef HAVE_VALGRIND_VALGRIND_H
 #include <valgrind/valgrind.h>
 #endif
 
+#if defined(_WIN64)
+#if defined(_M_ARM64)
+#define THUNK_SIZE 20
+typedef struct {
+  orc_uint32 FunctionLength : 18;
+  orc_uint32 Version : 2;
+  orc_uint32 HasExceptionHandler : 1;
+  orc_uint32 PackedEpilog : 1;
+  orc_uint32 EpilogCount : 5;
+  orc_uint32 CodeWords : 5;
+  orc_uint8  UnwindCode[4];
+  orc_uint32 ExceptionHandler;
+} UNWIND_INFO;
+#else // _M_X64
+#define THUNK_SIZE 12
+#define UWOP_SET_FPREG 3
+#define UWOP_PUSH_NONVOL 0
+#define UWOP_REG_RBP 5
+typedef union {
+  struct {
+    orc_uint8 CodeOffset;
+    orc_uint8 UnwindOp : 4;
+    orc_uint8 OpInfo : 4;
+  };
+  orc_uint16 FrameOffset;
+} UNWIND_CODE;
+typedef struct {
+  orc_uint8   Version : 3;
+  orc_uint8   Flags : 5;
+  orc_uint8   SizeOfProlog;
+  orc_uint8   CountOfCodes;
+  orc_uint8   FrameRegister : 4;
+  orc_uint8   FrameOffset : 4;
+  UNWIND_CODE UnwindCode[2];
+  orc_uint32  ExceptionHandler;
+} UNWIND_INFO;
+#endif // defined(_M_ARM64)
+typedef struct {
+  RUNTIME_FUNCTION function_table;
+  UNWIND_INFO unwind_info;
+  orc_uint8 thunk[THUNK_SIZE];
+} OrcUnwindInfo;
+#endif
+
 /**
  * SECTION:orccompiler
  * @title: OrcCompiler
@@ -32,14 +100,14 @@
  * or functions that use ORC_ASM_CODE() internally.
  */
 
-void orc_compiler_assign_rules (OrcCompiler *compiler);
-void orc_compiler_global_reg_alloc (OrcCompiler *compiler);
-void orc_compiler_rewrite_insns (OrcCompiler *compiler);
-void orc_compiler_rewrite_vars (OrcCompiler *compiler);
-void orc_compiler_rewrite_vars2 (OrcCompiler *compiler);
-int orc_compiler_dup_temporary (OrcCompiler *compiler, int var, int j);
-int orc_compiler_new_temporary (OrcCompiler *compiler, int size);
-void orc_compiler_check_sizes (OrcCompiler *compiler);
+static void orc_compiler_assign_rules (OrcCompiler *compiler);
+static void orc_compiler_global_reg_alloc (OrcCompiler *compiler);
+static void orc_compiler_rewrite_insns (OrcCompiler *compiler);
+static void orc_compiler_rewrite_vars (OrcCompiler *compiler);
+static void orc_compiler_rewrite_vars2 (OrcCompiler *compiler);
+static int orc_compiler_dup_temporary (OrcCompiler *compiler, int var, int j);
+static int orc_compiler_new_temporary (OrcCompiler *compiler, int size);
+static void orc_compiler_check_sizes (OrcCompiler *compiler);
 
 static char **_orc_compiler_flag_list;
 int _orc_compiler_flag_backup;
@@ -47,20 +115,104 @@ int _orc_compiler_flag_emulate;
 int _orc_compiler_flag_debug;
 int _orc_compiler_flag_randomize;
 
+/* For Windows */
+int _orc_codemem_alignment;
+
+#if defined(_WIN64)
+static DWORD orc_exception_handler(PEXCEPTION_RECORD exceptionRecord,
+                                   PEXCEPTION_REGISTRATION_RECORD _unused,
+                                   PCONTEXT context,
+                                   PEXCEPTION_REGISTRATION_RECORD *_unused2)
+{
+  return ExceptionContinueSearch;
+}
+#endif
+
 void
 _orc_compiler_init (void)
 {
-  const char *envvar;
+  char *envvar;
+#ifdef HAVE_CODEMEM_VIRTUALALLOC
+  #ifdef ORC_WINAPI_ONLY_APP
+    size_t page_size;
+  #else
+    size_t page_size ORC_GNUC_UNUSED;
+  #endif
 
-  envvar = getenv ("ORC_CODE");
+  SYSTEM_INFO info;
+#endif
+
+  envvar = _orc_getenv ("ORC_CODE");
   if (envvar != NULL) {
     _orc_compiler_flag_list = strsplit (envvar, ',');
+    free (envvar);
   }
 
   _orc_compiler_flag_backup = orc_compiler_flag_check ("backup");
   _orc_compiler_flag_emulate = orc_compiler_flag_check ("emulate");
   _orc_compiler_flag_debug = orc_compiler_flag_check ("debug");
   _orc_compiler_flag_randomize = orc_compiler_flag_check ("randomize");
+
+#ifdef HAVE_CODEMEM_VIRTUALALLOC
+  GetNativeSystemInfo(&info);
+  page_size = info.dwPageSize;
+
+  /* Protection attribute change via VirtualProtect will be applied per
+   * page memory unit. So we should split code memory with page aligned range.
+   * Otherwise the protection attribute of previously generated executable code
+   * memory can be affected by later generated one.
+   */
+  _orc_codemem_alignment = info.dwPageSize - 1;
+#else
+  /* 16 bytes alignment by default */
+  _orc_codemem_alignment = 15;
+#endif
+
+  int can_jit = TRUE;
+
+  if (!_orc_compiler_flag_backup && !_orc_compiler_flag_emulate) {
+    /* If backup code is not enabled and emulation is not enabled, that means
+     * we will do JIT compilation and call orc_code_region_allocate_codemem().
+     */
+#if defined(HAVE_CODEMEM_VIRTUALALLOC) && defined(ORC_WINAPI_ONLY_APP)
+    /* When targeting Windows Store apps, the codeGeneration capability must
+     * be enabled in the app manifest, or passing PAGE_EXECUTE to
+     * VirtualProtectFromApp will return NULL. In this case, we must force
+     * backup C code, and if that's not available, we must emulate. */
+    can_jit = FALSE;
+    void *mem = VirtualAllocFromApp (NULL, page_size, MEM_COMMIT,
+        PAGE_READWRITE);
+    if (mem) {
+      int old_protect;
+      if (VirtualProtectFromApp (mem, page_size, PAGE_EXECUTE, &old_protect) > 0)
+        can_jit = TRUE;
+      VirtualFree (mem, 0, MEM_RELEASE);
+    } else {
+      ORC_WARNING ("Unable to allocate executable pages: using backup code or "
+        "emulation: codeGeneration capability isn't set in the app manifest?");
+    }
+#elif defined(HAVE_CODEMEM_MMAP)
+    /* In this case, we need to check that we can mmap pages as executable.
+     * This is not the case under the combination of SELinux and sandboxing
+     * profiles.
+     */
+    can_jit = FALSE;
+    OrcCodeRegion *region = orc_code_region_alloc();
+    if (region) {
+      can_jit = TRUE;
+      // FIXME: the file descriptor should be kept somewhere
+      // Currently the underlying mmap'd pages are leaked
+      free(region);
+    } else {
+      ORC_WARNING ("Unable to allocate executable pages: using backup code or emulation");
+    }
+#endif
+
+    if (!can_jit) {
+      _orc_compiler_flag_backup = TRUE;
+      _orc_compiler_flag_emulate = TRUE;
+    }
+  }
 }
 
 int
@@ -76,7 +228,7 @@ orc_compiler_flag_check (const char *flag)
   return FALSE;
 }
 
-int
+static int
 orc_compiler_allocate_register (OrcCompiler *compiler, int data_reg)
 {
   int i;
@@ -96,8 +248,10 @@ orc_compiler_allocate_register (OrcCompiler *compiler, int data_reg)
     roff = rand()&0x1f;
   }
 
-  for(i=0;i<32;i++){
-    reg = offset + ((roff + i)&0x1f);
+  for (i = 0; i < ORC_N_REGS; i++) {
+    // For volatile registers, try up to 32 registers from offset
+    // Amyspark: this ensures that we don't run over vector registers
+    reg = offset + ((roff + i) & 0x1f);
     if (compiler->valid_regs[reg] &&
         !compiler->save_regs[reg] &&
         compiler->alloc_regs[reg] == 0) {
@@ -106,10 +260,14 @@ orc_compiler_allocate_register (OrcCompiler *compiler, int data_reg)
       return reg;
     }
   }
-  for(i=0;i<32;i++){
-    reg = offset + ((roff + i)&0x1f);
-    if (compiler->valid_regs[reg] &&
-        compiler->alloc_regs[reg] == 0) {
+  // Use ORC_N_REGS to ensure forward proofed iteration
+  for (i = 0; i < ORC_N_REGS; i++) {
+    // Try with up to 64 registers starting from offset
+    // If this lands us within vector range, bail out
+    reg = offset + ((roff + i) & 0x3f);
+    if (reg >= compiler->target->data_register_offset && !data_reg)
+      break;
+    if (compiler->valid_regs[reg] && compiler->alloc_regs[reg] == 0) {
       compiler->alloc_regs[reg]++;
       compiler->used_regs[reg] = 1;
       return reg;
@@ -117,7 +275,7 @@ orc_compiler_allocate_register (OrcCompiler *compiler, int data_reg)
   }
 
   if (data_reg || !compiler->allow_gp_on_stack) {
-    orc_compiler_error (compiler, "register overflow for %s reg",
+    orc_compiler_error (compiler, "register overflow for %s register",
         data_reg ? "vector" : "gp");
     compiler->result = ORC_COMPILE_RESULT_UNKNOWN_COMPILE;
   }
@@ -125,68 +283,37 @@ orc_compiler_allocate_register (OrcCompiler *compiler, int data_reg)
   return 0;
 }
 
-/**
- * orc_program_compile:
- * @program: the OrcProgram to compile
- *
- * Compiles an Orc program for the current CPU.  If successful,
- * executable code for the program was generated and can be
- * executed.
- *
- * The return value indicates various levels of success or failure.
- * Success can be determined by checking for a true value of the
- * macro ORC_COMPILE_RESULT_IS_SUCCESSFUL() on the return value.  This
- * indicates that executable code was generated.  If the macro
- * ORC_COMPILE_RESULT_IS_FATAL() on the return value evaluates to
- * true, then there was a syntactical error in the program.  If the
- * result is neither successful nor fatal, the program can still be
- * emulated.
- *
- * Returns: an OrcCompileResult
- */
-OrcCompileResult
-orc_program_compile (OrcProgram *program)
+#if defined(HAVE_CODEMEM_VIRTUALALLOC)
+static orc_bool
+_set_virtual_protect (void * mem, size_t size, int code_protect)
 {
-  return orc_program_compile_for_target (program, orc_target_get_default ());
+  char *msg;
+  DWORD old_protect;
+
+  /* No code, so we 'succeed' */
+  if (size == 0)
+    return TRUE;
+
+  if (!mem)
+    return FALSE;
+
+  if (_virtualprotect (mem, size, code_protect, &old_protect) > 0)
+    return TRUE;
+
+  FormatMessageA (FORMAT_MESSAGE_ALLOCATE_BUFFER
+      | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM, NULL,
+      GetLastError (), 0, (LPTSTR) &msg, 0, NULL);
+  ORC_ERROR ("Couldn't set memory protect on %p from %x to %x: %s", mem,
+      old_protect, code_protect, msg);
+  LocalFree (msg);
+
+  return FALSE;
 }
+#endif
 
-/**
- * orc_program_compile_for_target:
- * @program: the OrcProgram to compile
- *
- * Compiles an Orc program for the given target, using the
- * default target flags for that target.
- *
- * Returns: an OrcCompileResult
- */
 OrcCompileResult
-orc_program_compile_for_target (OrcProgram *program, OrcTarget *target)
+orc_compiler_compile_program (OrcCompiler *compiler, OrcProgram *program, OrcTarget *target, unsigned int flags)
 {
-  unsigned int flags;
-
-  if (target) {
-    flags = target->get_default_flags ();
-  } else {
-    flags = 0;
-  }
-
-  return orc_program_compile_full (program, target, flags);
-}
-
-/**
- * orc_program_compile_full:
- * @program: the OrcProgram to compile
- *
- * Compiles an Orc program for the given target, using the
- * given target flags.
- *
- * Returns: an OrcCompileResult
- */
-OrcCompileResult
-orc_program_compile_full (OrcProgram *program, OrcTarget *target,
-    unsigned int flags)
-{
-  OrcCompiler *compiler;
   int i;
   OrcCompileResult result;
   const char *error_msg;
@@ -209,9 +336,6 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
     program->asm_code = NULL;
   }
 
-  compiler = malloc (sizeof(OrcCompiler));
-  memset (compiler, 0, sizeof(OrcCompiler));
-
   if (program->backup_func) {
     program->code_exec = program->backup_func;
   } else {
@@ -223,17 +347,19 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
   compiler->target_flags = flags;
 
   {
-    ORC_LOG("variables");
+    ORC_LOG("Program variables");
     for(i=0;i<ORC_N_VARIABLES;i++){
       if (program->vars[i].size > 0) {
-        ORC_LOG("%d: %s size %d type %d alloc %d", i,
+        ORC_LOG("%d: %s size %d type %d alignment %d is_aligned %d alloc %d", i,
             program->vars[i].name,
             program->vars[i].size,
             program->vars[i].vartype,
+            program->vars[i].alignment,
+            program->vars[i].is_aligned,
             program->vars[i].alloc);
       }
     }
-    ORC_LOG("instructions");
+    ORC_LOG("Program instructions");
     for(i=0;i<program->n_insns;i++){
       ORC_LOG("%d: %s %d %d %d %d", i,
           program->insns[i].opcode->name,
@@ -272,23 +398,24 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
   orc_compiler_rewrite_vars (compiler);
   if (compiler->error) goto error;
 
-#if 0
   {
-    ORC_ERROR("variables");
+    ORC_LOG("Compiler variables");
     for(i=0;i<ORC_N_VARIABLES;i++){
       if (compiler->vars[i].size > 0) {
-        ORC_ERROR("%d: %s size %d type %d alloc %d [%d,%d]", i,
+        ORC_LOG("%d: %s size %d type %d alignment %d is_aligned %d alloc %d [%d,%d]", i,
             compiler->vars[i].name,
             compiler->vars[i].size,
             compiler->vars[i].vartype,
+            compiler->vars[i].alignment,
+            compiler->vars[i].is_aligned,
             compiler->vars[i].alloc,
             compiler->vars[i].first_use,
             compiler->vars[i].last_use);
       }
     }
-    ORC_ERROR("instructions");
+    ORC_LOG("Compiler instructions");
     for(i=0;i<compiler->n_insns;i++){
-      ORC_ERROR("%d: %s %d %d %d %d", i,
+      ORC_LOG("%d: %s %d %d %d %d", i,
           compiler->insns[i].opcode->name,
           compiler->insns[i].dest_args[0],
           compiler->insns[i].dest_args[1],
@@ -296,7 +423,6 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
           compiler->insns[i].src_args[1]);
     }
   }
-#endif
   program->orccode = orc_code_new ();
 
   program->orccode->is_2d = program->is_2d;
@@ -318,7 +444,7 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
     program->orccode->vars[i].value = compiler->vars[i].value;
   }
 
-  if (program->backup_func && _orc_compiler_flag_backup) {
+  if (program->backup_func && (_orc_compiler_flag_backup || target == NULL)) {
     orc_compiler_error (compiler, "Compilation disabled, using backup");
     compiler->result = ORC_COMPILE_RESULT_UNKNOWN_COMPILE;
     goto error;
@@ -355,10 +481,145 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
     goto error;
   }
 
+#if defined(_WIN64) && defined(ORC_SUPPORTS_BACKTRACE_FROM_JIT)
+  OrcUnwindInfo table;
+  // The structures must be DWORD aligned in memory.
+  const unsigned char *alignas_offset =
+      (unsigned char *)((DWORD64)(compiler->codeptr + 3) & (~3));
+
+  if (compiler->use_frame_pointer) {
+    memset(&table, 0, sizeof(OrcUnwindInfo));
+    const DWORD64 start_of_orcunwindinfo = alignas_offset - compiler->code;
+
+    table.function_table.BeginAddress = 0;
+
+    // The following Arm64 block is based on the Firefox changeset:
+    // https://hg.mozilla.org/mozilla-central/rev/4d932b82695c
+    // which I fixed to work with pure C.
+
+    // The program counter (and the stack pointer, on Arm64) must be modified
+    // for the handler to be recognised as valid.
+#ifdef _M_ARM64
+    table.function_table.UnwindData =
+        start_of_orcunwindinfo + ORC_STRUCT_OFFSET(OrcUnwindInfo, unwind_info);
+
+    // one 32-bit word gives us up to 4 codes
+    table.unwind_info.CodeWords = 1;
+    // alloc_s small stack of size 1*16
+    table.unwind_info.UnwindCode[0] = 0x1;
+    // end
+    table.unwind_info.UnwindCode[1] = 0xe4;
+
+    // xip0/r16 should be safe to clobber: Windows just used it to call the thunk.
+    const orc_uint8 reg = 16;
+
+    const void *handler = (void *)&orc_exception_handler;
+    const uint16_t *addr = (uint16_t *)&handler;
+
+    // Say `handler` is 0x4444333322221111, then:
+    table.thunk[0] = 0xd2800000 | addr[0] << 5 | reg;  // mov  xip0, 1111
+    table.thunk[1] = 0xf2a00000 | addr[1] << 5 | reg;  // movk xip0, 2222 lsl #0x10
+    table.thunk[2] = 0xf2c00000 | addr[2] << 5 | reg;  // movk xip0, 3333 lsl #0x20
+    table.thunk[3] = 0xf2e00000 | addr[3] << 5 | reg;  // movk xip0, 4444 lsl #0x30
+    table.thunk[4] = 0xd61f0000 | reg << 5;                // br xip0
+#else
+    table.function_table.EndAddress = compiler->codeptr - compiler->code;
+#ifdef __MINGW32__
+    // They're the same member, but MSYS2 only defines the former under x86.
+    table.function_table.UnwindData =
+#else
+    table.function_table.UnwindInfoAddress =
+#endif
+        start_of_orcunwindinfo + ORC_STRUCT_OFFSET(OrcUnwindInfo, unwind_info);
+
+    table.unwind_info.Version = 1;
+    // We handle the exception (let the UCRT terminate the app)
+    table.unwind_info.Flags = UNW_FLAG_EHANDLER;
+    table.unwind_info.SizeOfProlog = 8;
+    // See below
+    table.unwind_info.CountOfCodes = 2;
+    // RAX = 50, RBP = 55
+    // https://learn.microsoft.com/en-us/cpp/build/exception-handling-x64?view=msvc-170#operation-info
+    // https://www.felixcloutier.com/x86/mov
+    table.unwind_info.FrameRegister = UWOP_REG_RBP;
+    // No offset from RSP
+    table.unwind_info.FrameOffset = 0;
+    // Bytes 0-4 are the ENDBR64 CET instruction
+    // The docs kind of lie when they say that offsets are not allowed
+    // until SET_FPREG. I guess they apply to OpInfo instead?
+    table.unwind_info.UnwindCode[0].CodeOffset = 5;
+    table.unwind_info.UnwindCode[0].UnwindOp = UWOP_SET_FPREG;
+    table.unwind_info.UnwindCode[0].OpInfo = 0;  // mov rbp, rsp
+    table.unwind_info.UnwindCode[1].CodeOffset = 4;
+    table.unwind_info.UnwindCode[1].UnwindOp = UWOP_PUSH_NONVOL;
+    table.unwind_info.UnwindCode[1].OpInfo = UWOP_REG_RBP; // push rbp
+    // Same as above -- the exception handler is required, and must do
+    // something to actually be dispatched.
+    table.unwind_info.ExceptionHandler =
+        start_of_orcunwindinfo + ORC_STRUCT_OFFSET(OrcUnwindInfo, thunk);
+    // mov %rax, (address of handler as imm64)
+    // jmp %rax
+    static const orc_uint8 thunk[THUNK_SIZE] = {
+      0x48, 0xb8, 0, 0, 0, 0, 0, 0, 0, 0, 0xff, 0xe0,
+    };
+    memcpy (&table.thunk, &thunk, THUNK_SIZE);
+    memcpy (&table.thunk[2], &orc_exception_handler, 8);
+#endif
+
+    program->orccode->code_size = (unsigned char *)alignas_offset -
+                                  compiler->code + sizeof(OrcUnwindInfo);
+  } else {
+    program->orccode->code_size = compiler->codeptr - compiler->code;
+  }
+#else
   program->orccode->code_size = compiler->codeptr - compiler->code;
+#endif
+
   orc_code_allocate_codemem (program->orccode, program->orccode->code_size);
 
-  memcpy (program->orccode->code, compiler->code, program->orccode->code_size);
+#if defined(__APPLE__) && (!defined(TARGET_OS_OSX) || TARGET_OS_OSX)
+#if defined(MAC_OS_VERSION_11_0) && MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_VERSION_11_0
+  if (__builtin_available (macOS 11.0, *)) {
+    if (pthread_jit_write_protect_supported_np ())
+      pthread_jit_write_protect_np (0);
+  }
+#endif
+#endif
+#if defined(HAVE_CODEMEM_VIRTUALALLOC)
+  /* Ensure that code region is writable before memcpy */
+  _set_virtual_protect (program->orccode->code, program->orccode->code_size,
+      PAGE_READWRITE);
+#endif
+#if defined(_WIN64) && defined(ORC_SUPPORTS_BACKTRACE_FROM_JIT)
+  if (compiler->use_frame_pointer) {
+    void *const program_unwind_info =
+        program->orccode->code + (alignas_offset - compiler->code);
+    PRUNTIME_FUNCTION const runtime_function_address =
+          (PRUNTIME_FUNCTION)((orc_uint8*)program_unwind_info +
+                              ORC_STRUCT_OFFSET(OrcUnwindInfo, function_table));
+    const size_t real_code_size = compiler->codeptr - compiler->code;
+    memcpy (program->orccode->code, compiler->code, real_code_size);
+    memcpy (program_unwind_info, &table, sizeof (OrcUnwindInfo));
+    if (RtlAddFunctionTable (runtime_function_address, 1, (DWORD64)program->orccode->code)) {
+      DWORD64 dyn_base = 0;
+      PRUNTIME_FUNCTION const p = RtlLookupFunctionEntry (
+          (DWORD64)program->orccode->code + 20, &dyn_base, NULL);
+      if (p != runtime_function_address) {
+        ORC_ERROR ("Runtime function for program %s %p is bogus "
+                    "(dynbase=%llx, info=%p, expected=%p)",
+                    program->name, program->orccode->code, dyn_base, p,
+                    runtime_function_address);
+      }
+    } else {
+      ORC_WARNING ("Unable to install unwind info for program %s %p",
+                  program->name, program->orccode->code);
+    }
+  } else {
+    memcpy(program->orccode->code, compiler->code, program->orccode->code_size);
+  }
+#else
+  memcpy(program->orccode->code, compiler->code, program->orccode->code_size);
+#endif
 
 #ifdef VALGRIND_DISCARD_TRANSLATIONS
   VALGRIND_DISCARD_TRANSLATIONS (program->orccode->exec,
@@ -369,6 +630,23 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
     compiler->target->flush_cache (program->orccode);
   }
 
+#if defined(__APPLE__) && (!defined(TARGET_OS_OSX) || TARGET_OS_OSX)
+#if defined(MAC_OS_VERSION_11_0) && MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_VERSION_11_0
+  if (__builtin_available (macOS 11.0, *)) {
+    if (pthread_jit_write_protect_supported_np ()) {
+      pthread_jit_write_protect_np (1);
+      sys_icache_invalidate (program->orccode->exec, program->orccode->code_size);
+    }
+  }
+#endif
+#endif
+#if defined(HAVE_CODEMEM_VIRTUALALLOC)
+  /* Code region is now ready for execution */
+ if (!_set_virtual_protect (program->orccode->exec, program->orccode->code_size,
+       PAGE_EXECUTE))
+   /* Can't set code as executable, force emulation */
+   program->orccode->exec = (void *)orc_executor_emulate;
+#endif
   program->code_exec = program->orccode->exec;
 
   program->asm_code = compiler->asm_code;
@@ -416,7 +694,7 @@ error:
   return result;
 }
 
-void
+static void
 orc_compiler_check_sizes (OrcCompiler *compiler)
 {
   int i;
@@ -535,7 +813,7 @@ get_store_opcode_for_size (int size)
   return NULL;
 }
 
-void
+static void
 orc_compiler_rewrite_insns (OrcCompiler *compiler)
 {
   int i;
@@ -565,7 +843,7 @@ orc_compiler_rewrite_insns (OrcCompiler *compiler)
         if (var->vartype == ORC_VAR_TYPE_SRC ||
             var->vartype == ORC_VAR_TYPE_DEST) {
           OrcInstruction *cinsn;
-          
+
           cinsn = compiler->insns + compiler->n_insns;
           compiler->n_insns++;
 
@@ -636,7 +914,7 @@ orc_compiler_rewrite_insns (OrcCompiler *compiler)
         var = compiler->vars + insn.dest_args[i];
         if (var->vartype == ORC_VAR_TYPE_DEST) {
           OrcInstruction *cinsn;
-          
+
           cinsn = compiler->insns + compiler->n_insns;
           compiler->n_insns++;
 
@@ -654,7 +932,7 @@ orc_compiler_rewrite_insns (OrcCompiler *compiler)
   }
 }
 
-void
+static void
 orc_compiler_assign_rules (OrcCompiler *compiler)
 {
   int i;
@@ -705,7 +983,8 @@ orc_compiler_get_temp_reg (OrcCompiler *compiler)
   ORC_DEBUG("at insn %d %s", compiler->insn_index,
       compiler->insns[compiler->insn_index].opcode->name);
 
-  for(j=compiler->min_temp_reg;j<ORC_VEC_REG_BASE+32;j++){
+  // Use ORC_N_REGS to ensure forward proofed iteration
+  for (j = compiler->min_temp_reg; j < ORC_N_REGS; j++) {
     if (compiler->valid_regs[j] && !compiler->alloc_regs[j]) {
       compiler->min_temp_reg = j+1;
       if (compiler->max_used_temp_reg < j) compiler->max_used_temp_reg = j;
@@ -719,7 +998,7 @@ orc_compiler_get_temp_reg (OrcCompiler *compiler)
   return 0;
 }
 
-void
+static void
 orc_compiler_rewrite_vars (OrcCompiler *compiler)
 {
   int j;
@@ -831,7 +1110,7 @@ orc_compiler_rewrite_vars (OrcCompiler *compiler)
   }
 }
 
-void
+static void
 orc_compiler_global_reg_alloc (OrcCompiler *compiler)
 {
   int i;
@@ -895,15 +1174,16 @@ orc_compiler_global_reg_alloc (OrcCompiler *compiler)
 
   if (compiler->alloc_loop_counter && !compiler->error) {
     compiler->loop_counter = orc_compiler_allocate_register (compiler, FALSE);
-    /* FIXME massive hack */
-    if (compiler->loop_counter == 0) {
+    // FIXME: If loop_counter is invalid, the counter must be set manually
+    // in the executor 
+    if (compiler->loop_counter == ORC_REG_INVALID) {
       compiler->error = FALSE;
       compiler->result = ORC_COMPILE_RESULT_OK;
     }
   }
 }
 
-void
+static void
 orc_compiler_rewrite_vars2 (OrcCompiler *compiler)
 {
   int i;
@@ -969,7 +1249,7 @@ orc_compiler_rewrite_vars2 (OrcCompiler *compiler)
 
 }
 
-int
+static int
 orc_compiler_dup_temporary (OrcCompiler *compiler, int var, int j)
 {
   int i = ORC_VAR_T1 + compiler->n_temp_vars + compiler->n_dup_vars;
@@ -983,7 +1263,7 @@ orc_compiler_dup_temporary (OrcCompiler *compiler, int var, int j)
   return i;
 }
 
-int
+static int
 orc_compiler_new_temporary (OrcCompiler *compiler, int size)
 {
   int i = ORC_VAR_T1 + compiler->n_temp_vars + compiler->n_dup_vars;
@@ -997,11 +1277,13 @@ orc_compiler_new_temporary (OrcCompiler *compiler, int size)
   return i;
 }
 
-void
+#if 0
+static void
 orc_compiler_dump_asm (OrcCompiler *compiler)
 {
   printf("%s", compiler->asm_code);
 }
+#endif
 
 /**
  * orc_compiler_append_code:
@@ -1041,14 +1323,14 @@ orc_compiler_label_new (OrcCompiler *compiler)
   return compiler->n_labels++;
 }
 
-void
+static void
 orc_compiler_load_constant (OrcCompiler *compiler, int reg, int size,
     int value)
 {
   compiler->target->load_constant (compiler, reg, size, value);
 }
 
-void
+static void
 orc_compiler_load_constant_long (OrcCompiler *compiler, int reg,
     OrcConstant *constant)
 {
@@ -1199,16 +1481,23 @@ orc_compiler_get_constant_reg (OrcCompiler *compiler)
 
 #define ORC_COMPILER_ERROR_BUFFER_SIZE 200
 
-void
+static void
 orc_compiler_error_valist (OrcCompiler *compiler, const char *fmt,
     va_list args)
 {
-  char *s;
+  char *s = NULL;
 
   if (compiler->error_msg) return;
 
+#ifdef HAVE_VASPRINTF
+  vasprintf (&s, fmt, args);
+#elif defined(_UCRT)
   s = malloc (ORC_COMPILER_ERROR_BUFFER_SIZE);
-  vsprintf (s, fmt, args);
+  vsnprintf_s (s, ORC_COMPILER_ERROR_BUFFER_SIZE, _TRUNCATE, fmt, args);
+#else
+  s = malloc (ORC_COMPILER_ERROR_BUFFER_SIZE);
+  vsnprintf (s, ORC_COMPILER_ERROR_BUFFER_SIZE, fmt, args);
+#endif
   compiler->error_msg = s;
   compiler->error = TRUE;
   compiler->result = ORC_COMPILE_RESULT_UNKNOWN_COMPILE;
@@ -1224,3 +1513,36 @@ orc_compiler_error (OrcCompiler *compiler, const char *fmt, ...)
   va_end (var_args);
 }
 
+void
+orc_compiler_emit_invariants (OrcCompiler *compiler)
+{
+  int j;
+  OrcInstruction *insn;
+  OrcStaticOpcode *opcode;
+  OrcRule *rule;
+
+  for(j=0;j<compiler->n_insns;j++){
+    insn = compiler->insns + j;
+    opcode = insn->opcode;
+
+    if (!(insn->flags & ORC_INSN_FLAG_INVARIANT)) continue;
+
+    ORC_ASM_CODE(compiler,"# %d: %s\n", j, opcode->name);
+
+    compiler->insn_shift = compiler->loop_shift;
+    if (insn->flags & ORC_INSTRUCTION_FLAG_X2) {
+      compiler->insn_shift += 1;
+    }
+    if (insn->flags & ORC_INSTRUCTION_FLAG_X4) {
+      compiler->insn_shift += 2;
+    }
+
+    rule = insn->rule;
+    if (rule && rule->emit) {
+      rule->emit (compiler, rule->emit_user, insn);
+    } else {
+      orc_compiler_error (compiler, "no code generation rule for %s",
+          opcode->name);
+    }
+  }
+}
